	PROGRAM EFLUXPROG

C            VERSION 2             5 MARCH 2002
C
C    Address questions and comments about the routines in this file to
C
C         Gordon D. Holman
C         NASA/Goddard Space Flight Center
C         Code 682
C         Greenbelt, MD 20771
C         Office phone: (301) 286-4636
C		E-Mail: holman@stars.gsfc.nasa.gov
C
C    Address questions and comments about SUBROUTINE ELECTRONFLUX and 
C	its subprograms to
C
C		James M. McTiernan
C		University of California, Berkeley
C		Space Science Laboratory #7450
C		Berkeley, CA 94720-7450
C		Office phone: (510) 643-9246
C		E-Mail: jimm@ssl.berkeley.edu
C
C    Purpose -
C         The program computes the steady-state electron distribution 
C         function in a flare magnetic loop as a function of position
C		along a magnetic field line, electron pitch angle, and
C         electron energy.  Subroutine ELECTRONFLUX computes the 
C		distribution function everywhere along the field line, given
C		the injected flux at the top of the field line and a model
C		solar atmosphere.  The steady-state Fokker-Planck equation
C		is integrated.  Coulomb and synchrotron energy losses and 
C		pitch angle diffusion are included, as well as magnetic 
C		mirroring.  
C		
C    Files used -  
C		efluxprog.f			Main Program & related subroutines - this file
C		efluxsubs.f			Subroutines called by Subroutine ELECTRONFLUX
C		electronflux.f		Subroutine ELECTRONFLUX - Computes the electron flux
C		ssfpinc.f			Include file - common blocks, parameter definitions
C		efluxin.txt			Input data file, read by Subroutine FPINPUT
C
C		efluxout.txt		Output file, written by Subroutine FPOUTPUT
C
C    Documentation - 
C         efluxdoc.pdf
C
C    Revisions -
C		Version 1: 2 April 2001
C		Version 2: 5 March 2002 - Conversions from density to flux and back corrected 
C									by replacing gammav(i) with v(i).

	INCLUDE 'SSFPINC.F'

	INTEGER TIME0, ITIME, ITICK

      CALL SYSTEM_CLOCK(TIME0)

C	*****************************************************
C	*                 INPUT PARAMETERS                  *
C	*****************************************************
C	INJECTED ELECTRON ENERGY DISTRIBUTION
C				Double power law
C
C	Lower (plow) and upper (phigh) power-law indices
	plow  = 5.
	phigh = 5.
C	Lower cutoff, break, and upper cutoff energies (keV)
	emin = 15.
	ebrk = 200.
	emax = 1.0E+06
C	Number density of injected electrons (cm^-3)
	dniel = 2.0E+08
C	*****************************************************
C	INJECTED ELECTRON PITCH ANGLE DISTRIBUTION
C
C	kfan = 0: all electrons have 0 degree pitch angle
C	kfan = 1: isotropic electron distribution
C	kfan = 2: isotropic in the forward hemisphere
C
	kfan = 1
C	*****************************************************
C	ATMOSPHERIC DENSITY HEIGHT DISTRIBUTIONS
C
C	The provided model atmosphere has a constant density 
C	corona above h=0, and the density increases exponentially 
C	below h=0.  Heights and densities that characterize the 
C	atmosphere are read in subroutine fpinput from the input file
C	efluxin.txt.  The exponential growth scale length, hexp, 
C	is computed from the densities specified for electrons.  
C	*****************************************************
C	MAGNETIC FIELD AND MAGNETIC MIRRORING
C
C	khs = 0  no magnetic mirroring, B is constant (= bmin)
C	khs = 1  magnetic mirroring is enabled
C
	khs = 1
C
C	ksc = 0  the coronal magnetic field is vertical
C	ksc = 1  the coronal magnetic field is semicircular
C
	ksc = 1
C
C	kab=0 & khs=1: B increases linearly between ybmin and ybmax
C	kab=1 & khs=1: B increases exponentially between ybmin and ybmax
C
	kab = 0
C
C	Minimum and maximum values of y over which the magnetic
C	field is allowed to vary, in kilometers
C
	ybmin = 0.
	ybmax = 33570.
C
C	Magnetic field strength at ybmin and ybmax, in Gauss
C
	bmin = 200.
	bmax = 1000.
C	*****************************************************
C	Set FLAGS for PHYSICAL PROCESSES to include or exclude
C	 0 = No (Exclude), 1 = Yes (Include)
C
C	Include Coulomb energy losses?
	kb = 1
C	Include Coulomb pitch angle diffusion?
	kb1 = 1
C	Include synchrotron energy losses and diffusion?
	ksy = 1
C	Include synchrotron pitch angle diffusion?
	ksmu = 1
C
C	Include reflection of electrons at the top of the loop?
	kbc = 0
C	*****************************************************
C	COMPUTATION CONTROLS
C
C	Set ktrm = 0 to allow unphysical negative numbers to be
C	removed while solving for FLUX
	ktrm = 0
C	Number of iterations allowed per energy
	kitm = 20
C	Number of iterations until divergence check
	kkr = 3
C	Value of tau/eta beyond which solution is set to zero
	x1 = 30.
C	Relative accuracy desired
	adp = 3.E-07
C	Relative accuracy required
	adpm = 5.E-01
C	*****************************************************
C	*****************************************************

C Read additional input data from efluxin.txt

	call fpinput()

C Set up the x (pitch angle), y (position), z (energy) coordinate mesh

	call mesh()

C Align the coordinates for the atmosphere as a function of height
C with the coordinates for the electron distribution as a function
C of position along the magnetic field line

	call coordmap()

C Calculate the atmospheric density of each constituent as a function
C of position.  

	call atmosphere()

C Calculate the magnetic field strength as a function of position

	call magfield()

C Set up the injected electron flux distribution

	call fluxin()

C Compute the electron flux distribution at all grid points

      call electronflux(kmax,x,jmax,y,imax,z,nc,za,zn,ei,dn,bf,
     &     ftop,flux,kb,kb1,khs,kbc,ksy,ksmu,ktrm,kitm,kkr,x1,adp,
     &     adpm)

C Print out the results, including the input data, to efluxtout.txt

	call fpoutput()

      CALL SYSTEM_CLOCK(ITIME, ITICK)
      IF (ITICK /= 0)                                                  
     1WRITE (6,'(" Processing complete in ", F9.3, " seconds")')
     2  REAL(ITIME-TIME0)/REAL(ITICK)

	end

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

	subroutine atmosphere()
	include 'ssfpinc.f'

C	Calculates the atmospheric densities of the constituents
C	as a function of height and y(j).  Also computes the total density
C	and column depth as a function of y(j).  

C	*****************************************************
C	MODEL ATMOSPHERE:
C	Constant density corona, density increases 
C	exponentially below h = 0.  
C	*****************************************************

C	Calculate the exponential growth scale length 
C	for the constituent densities below h = 0.
C	The electron densities at h=0 and at h(nh) are used
C	to specify a common scale length for all constituents.

	   if(dni(3,nh) .eq. dni(3,nh-1)) then
	      hexp = 1.0E+35
	   else
	      hexp = -h(nh)/alog(dni(3,nh)/dni(3,nh-1))
	   end if

C	For each constituent
      do l=1,nc

C        Be sure that dni is not zero, min dni = 1.0e-10
         do i=1,nh
            if(dni(l,i).lt.1.0e-10) dni(l,i)=1.0e-10
         end do

C	   Compute the densities dn(hkm)
         do j=1,jmp1
            dn(l,j)=0.0
		  if(hkm(j) .ge. 0.00) then
	        dn(l,j) = dni(l,1)
	      else
			temp = -hkm(j)/hexp
              dn(l,j)= dni(l,1)*exp(temp)
	      end if
         end do

	end do

C	   noc(j) is the density in hydrogen masses at each point y(j)

      do j=1,jmp1
         noc(j)=0.0

         do l=1,nc
	     noc(j)=noc(j)+(dn(l,j)*zn(l))
         end do

C	   coldp(j) is the column depth in cm**-2 to each point y(j).
C	   Accounts for constant and exponential density distrubutions.

         if(j.eq.1) then
            coldp(j) = 0.0
         else
            j1=j-1
            if(noc(j).eq.noc(j1)) then
               coldp(j)=coldp(j1)+noc(j)*dy(j)
            else
               coldp(j)=coldp(j1)+(noc(j)-noc(j1))*hexp*1.E+05
            endif
         endif
	end do

	return
	end

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

	subroutine coordmap()
	include 'ssfpinc.f'

C	The atmosphere is stratified into evenly spaced horizontal layers. 
C	y(j) is the depth along the field lines of the top of each 
C	segment j, in cm, corresponding to each height hkm(j).
C	Align hkm(jmax+1) with y(jmax+1), with h(nh) in cms.
C	The magnetic field curvature is needed to do this.
C
C	schi is the sin of the angle (chi) between the magnetic field and 
C	horizontal.
C	thetab is the angle between the magnetic field and the vertical
C     (pi/2 - chi) in radians

      jmp1=jmax+1
      hkm(jmp1)=h(nh)

      if(ksc .eq. 1) then
C	   ksc = 1: semi-circular loop above hkm = 0., vertical below hkm = 0
C	            rkm is the radius (and height) of the loop
C				ydy(jy-2) is the distance the electrons travel in the
C				   corona above hkm = 0

	   rkm = ydy(jy-2)*(1.0e-5)/pieby2

         do j=1,jmax
            l=jmp1-j
		  if(y(l) .ge. ydy(jy-2)) then
			schi(l) = 1.0
			thetab(l) = 0.0
			hkm(l) = hkm(l+1)+dy(l)*(1.0e-5)
		  else
			schi(l) = sin(y(l)*(1.0e-5)/rkm)
			thetab(l) = pieby2 - y(l)*(1.0e-5)/rkm
			hkm(l) = rkm*cos(y(l)*(1.0e-5)/rkm)
		  endif
	   end do
      else
C	   ksc = 0: coronal loop is vertical
         do j=1,jmax
		  l=jmp1-j
            schi(l)=1.0
            thetab(l)=0.0
		  hkm(l)=hkm(l+1)+dy(l)*(1.0e-5)
	   end do
      endif

	return
	end

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

	subroutine edistin()
	include 'ssfpinc.f'

	real denoma, denomb, denom, K
	real zmin, zbrk, zmax

C Provide the injected electron distribution at y = 0, fz(i), 
C as a function of energy, z(i) (electrons/cc/keV).

C DOUBLE POWER LAW ENERGY DISTRIBUTION

C	Normalize energies to electron rest mass energy
	zmin = emin/mc2
	zbrk = ebrk/mc2
	zmax = emax/mc2

C	Normalize so that the integral of the distrubution function over energy
C	gives the injected electron density (dniel)

	  denoma = (zbrk**(1.0-plow))-(zmin**(1.0-plow))
	  denoma = denoma*(1.0-phigh)
	  denomb = (zmax**(1.0-phigh))-(zbrk**(1.0-phigh))
	  denomb = denomb*(1.0-plow)*(zbrk**(phigh-plow))
	  denom  = denoma + denomb
	  K = (1.0-phigh-plow+(plow*phigh))/denom

	do i = 1, imax
		if (z(i) .GE. zmin .AND. z(i) .LE. zbrk) then
			fz(i) = z(i)**(-plow)
		else if (z(i) .GT. zbrk .AND. z(i) .LE. zmax) then
			fz(i) = (zbrk**(phigh-plow))*z(i)**(-phigh)
		else
			fz(i) = 1.0E-30
		endif
	  fz(i) = dniel*K*fz(i)/mc2
	end do

	return
	end

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

	subroutine fluxin()
	include 'ssfpinc.f'

	call edistin()
	call padistin()

C Make ftop the input array at the top for each energy

      do i=1,imax
C	   Convert from density distribution to flux distribution
	   gamma(i) = z(i)+1.
	   v(i) = sqrt(1.-(1./gamma(i)))*spdl
         do k=1,kmax
            ftop(i,k)=fz(i)*xfm(k)*v(i)
         end do
	end do

	return
	end

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

	subroutine fpinput()

	include 'ssfpinc.f'

C	Read the input file 'efluxin.txt'

      open(unit=10,file='efluxin.txt', readonly, status='old')

C	read in the properties of the atmospheric constituents

      read(10,'(2i5)') nc,nh
      do l=1,nc
         read(10,'(3f12.4)') zn(l),za(l),ei(l)
	end do
C
C	read in the model atmosphere critical heights and densities

      do l=1,nh
         read(10,'(f12.4,5e12.4)') h(l),(dni(l1,l),l1=1,nc)
	end do
C
C	Read in the meshes (pitch angle cos, position, energy)

      read(10,'(3i5)') kx,jy,iz
      read(10,'(6f12.4)')(x(k),k=1,kx)
      read(10,'(6f12.4)')(ydy(j),j=1,jy)
      read(10,'(6f12.4)')(zdz(i),i=1,iz)

C close the input file

      close(unit=10)

	return
	end

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

	subroutine fpoutput()
	include 'ssfpinc.f'
      open(unit=19,file='efluxout.txt')
      write(19,'(15i5)') nc,nh,kfan,kitm,kkr
      write(19,'(15i5)') kb,kb1,khs,ksc,kab,kbc,kx,jy,iz
      write(19,'(15i5)') ksy,ksmu
      write(19,'(1p6e12.4)') sh,adp,adpm
C	write out electron distribution function input parameters
      write(19,'(1p6e12.4)') plow,phigh,emin,ebrk,emax,dniel
C	write out magnetic field input parameters
      write(19,'(1p6e12.4)') ybmin, ybmax, bmin, bmax, dbdy, ybexp
C
      write(19,'(3f12.6)') fzmc,zinc,mc2
      write(19,'(4f12.6)') x1
C	atmospheric constituents and densities
      do l=1,nc
         write(19,'(3f12.4)') zn(l),za(l),ei(l)
	end do
      do l=1,nh
         write(19,'(f12.4,1p5e12.4)') h(l),(dni(l1,l),l1=1,nc)
	end do
C	write out the meshes as input
      write(19,'(1p6e12.4)')(x(k),k=1,kx)
      write(19,'(1p6e12.4)')(ydy(j),j=1,jy)
      write(19,'(1p6e12.4)')(zdz(i),i=1,iz)
C	Now write out the stuff that was calculated
      write(19,'(15i5)') imax,jmax,kmax
      write(19,'(6f12.4)')(pa(k)*57.296,k=1,kmax)
      write(19,'(1p6e12.4)')(y(j),j=1,jmp1)
      write(19,'(1p6e12.4)')(z(i)*511.,i=1,imax)
      write(19,'(1p6e12.4)')(thetab(j),j=1,jmax)
      write(19,'(1p6e12.4)')(schi(j),j=1,jmax)
      write(19,'(1p6e12.4)')(hkm(j),j=1,jmp1)
      do j=1,jmp1
         write(19,'(1p6e12.4)') hkm(j),(dn(l1,j),l1=1,nc)
	end do
      write(19,'(1p6e12.4)')(noc(j),j=1,jmp1)
      write(19,'(1p6e12.4)')(coldp(j),j=1,jmp1)
      write(19,'(1p6e12.4)')(bf(j),j=1,jmp1)
      write(19,'(1p6e12.4)')(fz(i),i=1,imax)
      write(19,'(1p6e12.4)')(xfm(k),k=1,kmax)
C	Convert from electron flux to electron density
	do j=1,jmax
	   do k=1,kmax
	      do i=1,imax
	         flux(j,k,i)=flux(j,k,i)*511000.0/v(i)
		  end do
	   end do
	end do
C	Write electron distribution function (electrons/cc/keV)
      do j=1,jmax
         do k=1,kmax
            write(19,'(1p6e12.4)')(flux(j,k,i),i=1,imax)
         end do
	end do

	return
      end

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

	subroutine magfield()
	include 'ssfpinc.f'

C	MAGNETIC FIELD MODELS
C	if khs = 0  the d(ln B)/ds terms are switched off
C	if khs = 1  magnetic mirroring is enabled
C
C	kab=0: bf increases linearly between ybmin and ybmax
C		with slope dbdy (Gauss/km)
C	kab=1: bf increases exponentially between ybmin and ybmax
C		with scale length ybexp (km) 
C		d(ln B)/ds = const = 1/ybexp

	integer jb

	dbdy  = 0.
	ybexp = 0.

C	convert to cm
	ybmin = ybmin * 1.0E+05
	ybmax = ybmax * 1.0E+05

	if(khs .eq. 0) then
	  do jb=1,jmp1
	    bf(jb) = bmin
	  end do

	else

	  if(kab .eq. 0) then

	    dbdy = (bmax-bmin)/(ybmax-ybmin)

	    do jb=1,jmp1
	      if(y(jb) .lt. ybmin) then
	        bf(jb) = bmin
	      else if((y(jb) .ge. ybmin) .and. (y(jb) .le. ybmax)) then

	        bf(jb) = bmin + dbdy * (y(jb)-ybmin)

	      else if(y(jb) .gt. ybmax) then
	        bf(jb) = bmax
	      end if
	    end do

        else if(kab .eq. 1) then

	    ybexp = (ybmax-ybmin)/alog(bmax/bmin)

	    do jb=1,jmp1
	      if(y(jb) .lt. ybmin) then
	        bf(jb) = bmin
	      else if((y(jb) .ge. ybmin) .and. (y(jb) .le. ybmax)) then

	        bf(jb) = bmin * exp((y(jb)-ybmin)/ybexp)

	      else if(y(jb) .gt. ybmax) then
	        bf(jb) = bmax
	      end if
	    end do
	  end if
	end if

C	convert back to km
	ybmin = ybmin * 1.0E-05
	ybmax = ybmax * 1.0E-05
	dbdy = dbdy * 1.0E+05
	ybexp = ybexp * 1.0E-05

	return
	end

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

	subroutine mesh()

	include 'ssfpinc.f'

C	Form the x,y,z mesh

C	PITCH ANGLE
C	x = cos( pitch angle )

      kmax = kx
      do k=1,kmax
         pa(k) = acos(x(k))
	end do

C	POSITION
C	Convert km to cm

      do j=1,jy
         ydy(j)=ydy(j)*1.0e+5
	end do

C	The ydy with odd indices [ydy(1), ydy(3), etc.] are positions
C	The ydy with even indices [ydy(2), ydy(4), etc.] are increments
C	If jy = ndy, all the ydy are positions

      if(jy.eq.ndy) then
         do j=1,jy
	    y(j)=ydy(j)
	   end do
         jmax=jy
         y(jy+1)=2.0*y(jy)-y(jy-1)	
      else
         k1=1
         do 300 l=2,jy,2
	    y(k1)=ydy(l-1)
	    l1=.99+(ydy(l+1)-ydy(l-1))/ydy(l)
	    do 300 n=1,l1
               k=k1+1
               y(k)=y(k1)+ydy(l)
               k1=k
 300     continue
         jmax=k-1
         y(jmax+1)=ydy(jy)
      endif
      jmp1=jmax+1
      do j=1,jmax
         l=jmp1-j
         dy(l)=y(l+1)-y(l)
	end do

C	ENERGY
C	Electron energy (MeV)
C	The zdz with odd indices [zdz(1), zdz(3), etc.] are energies
C	The zdz with even indices [zdz(2), zdz(4), etc.] are increments

      k1=1
      do 310 l=2,iz,2
         z(k1)=zdz(l-1)
         l1=.99-((zdz(l-1)-zdz(l+1))/zdz(l))
         do 310 n=1,l1
            k=k1+1
            z(k)=z(k1)+zdz(l)
            k1=k
 310  continue
      imax=k
      z(k)=zdz(iz)
C	z's in units of mc2
C	zinc = 1 / 0.511
      do i=1,imax
         z(i) = z(i)*zinc
	end do

C	PITCH ANGLE: Determine index for which pa first exceeds 90 degrees
C	This needs to be done here since k1 et al., are constants
      do k=1,kmax
         if(x(k).gt.0.0) then 
            k0=k
         else
            if(x(k).eq.0.0) then
               k1=k
               kmax=kmax-1
               do kk=k1,kmax
                  x(kk)=x(kk+1)
			 end do
            else
               if(k0.eq.(k-1)) k1=k
            endif
         endif
	end do
      km=kmax-1

	return
	end

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

	subroutine padistin()

C Specify the pitch angle dependence of the injected electron 
C distribution function.

	include 'ssfpinc.f'

C kfan = 0: all electrons have 0 degree pitch angle
C kfan = 1: isotropic electron distribution
C kfan = 2: isotropic in the forward hemisphere

      if(kfan.eq.0) then
		xfm(1) = 1.0
	    do k=2,kmax
               xfm(k) = 0.0
	    end do
      else if(kfan.eq.1) then
         do k=1,kmax
	    xfm(k) = 1.0
         end do
      else if(kfan.eq.2) then
	    xxx=39.00/pieby2
	    do k=1,k0
               xfm(k)=1.0
		end do
	    do k=k1,km
               temp=xxx*(pa(k)-pieby2)
               xfm(k)=10.0**(-temp)
		end do
         xfm(1)=1.0
         xfm(kmax)=0.0
      else
         write(6,*) 'Bad value of kfan'
      endif

	return
	end

C23456789012345678901234567890123456789012345678901234567890123456789012